package at.tugraz.kmi.medokyservice.fca.db.domainmodel;

import java.util.HashSet;
import java.util.Set;

import at.tugraz.kmi.medokyservice.fca.db.DataObject;
import at.tugraz.kmi.medokyservice.fca.db.usermodel.User;

/**
 * Class representing a Knowledge Domain. Every Domain has an owner an {@link IncidenceMatrix} and a {@link Lattice}
 * computed from the IncidenceMatrix.
 * 
 * @author Bernd Prünster <mail@berndpruenster.org>
 * 
 */
public class Domain extends DataObject {

  /**
   * 
   */
  private static final long serialVersionUID = -9169088905113687870L;
  private IncidenceMatrix   mapping;
  private Lattice           formalContext;
  private Set<User>         owners;
  private boolean           global;
  private boolean           approved;

  /**
   * Creates a new domain form an {@link IncidenceMatrix} computing a {@link Lattice}.
   * 
   * @param name
   *          The domain name
   * @param description
   *          the domain description
   * @param matrix
   *          the {@link IncidenceMatrix} the domain is based upon.
   * @param owner
   *          the domain owner
   */
  public Domain(String name, String description, IncidenceMatrix matrix, User owner, boolean global) {
    super(name, description);
    mapping = matrix;
    formalContext = new Lattice(matrix);
    owners = new HashSet<User>();
    owners.add(owner);
    this.global = global;
    this.approved = false;
  }

  public Domain(String name, String description, IncidenceMatrix matrix, Set<User> owner, boolean global) {
    super(name, description);
    mapping = matrix;
    formalContext = new Lattice(matrix);
    owners = new HashSet<User>();
    owners.addAll(owner);
    this.global = global;
    this.approved = false;
  }

  /**
   * Returns the {@link IncidenceMatrix} the domain is based upon
   * 
   * @return The {@link IncidenceMatrix}
   */
  public IncidenceMatrix getMapping() {
    return mapping;
  }

  /**
   * Returns the {@link Lattice} of this domain
   * 
   * @return the lattice of the domain
   */
  public Lattice getFormalContext() {
    return formalContext;
  }

  public Set<User> getOwners() {
    return owners;
  }

  public void setOwners(Set<User> owners) {
    this.owners = owners;
  }

  public boolean isGlobal() {
    return global;
  }

  public boolean isApproved() {
    return approved;
  }

  public void setApproved(boolean approved) {
    this.approved = approved;
  }

  public void setMetadata() {
    mapping.getAttributes();
  }

  public void updateLattice() {
    formalContext = new Lattice(getMapping());
  }

}
